import torch
from torchtext import data
from torchtext import datasets
import random
import time
import os
import copy
import csv
import numpy as np
import torch.nn.utils.rnn as rnn_utils
from torch.utils.data.sampler import SubsetRandomSampler

path = os.path.join("..", "Coverage")                                    #PATH TO GAME FOLDERS, MODIFY 
game_lower_index = 31
game_upper_index = 130

class zork_dataset_steps(torch.utils.data.Dataset):
    
    def __init__(self, folder_path):
        self.path = folder_path
        ds_path = os.path.join(folder_path, "665_data.txt")
        ds_f = open(ds_path, "r")
        ds_lines = ds_f.readlines()
        self.game_ns = []
        self.step_ns = []
        for x in ds_lines:
            xsp = x.split()
            
            #info for retrieval
            tmp_game = xsp[0][4:]
            #for i in range(int(xsp[1])+1,int(xsp[1])+1):
            for i in range(500,501):
                self.game_ns.append(tmp_game)
                self.step_ns.append(str(i))
                
                #parse and store labels
                full_path = os.path.join(folder_path, xsp[0], "Step"+str(i))

                info_paths = []
                info_paths.append(os.path.join(full_path, "rooms.c.gcov")) # CHANGE if i want more files
                info_paths.append(os.path.join(full_path, "nrooms.c.gcov")) # CHANGE if i want more files
                label = self.get_coverage(info_paths) # grab coverage information from files
                coverage_score = 0.0
                for i in label:
                    if i > 0.5:
                        coverage_score += 1.0
                coverage_denom = max(1, len(label))
                coverage_score = coverage_score / coverage_denom
                
                #print(coverage_score)
                output_validity = self.parse_output(os.path.join(full_path, "ZorkOut.gameOut"))
                #print(output_validity)
                
                full_label = []
                full_label.append(1.0)
                full_label.append(output_validity)
                full_label.append(coverage_score)
                
                label_path = os.path.join(full_path, "label.csv")
                with open(label_path, 'w', newline='') as myfile:
                    wr = csv.writer(myfile)
                    wr.writerow(full_label)

        self.len = len(self.step_ns)
        
    def __getitem__(self, index):
        item_path = os.path.join(self.path, "Game"+self.game_ns[index], "Step"+self.step_ns[index])
        input_path = os.path.join(item_path, "input")
        label_path = os.path.join(item_path, "label.csv")
        x = np.loadtxt(input_path, delimiter='\n', dtype='str')
        x = x.tolist()
        y = np.loadtxt(label_path, delimiter=',', dtype=np.float32)  
        return x, y
    
    def __len__(self):
        return self.len
    
    def get_coverage(self, filepaths):
        label = []
        for filepath in filepaths:
            info_f = open(filepath, "r")
            info_lines = info_f.readlines()
            i = 0
            for y in info_lines:
                i += 1
                temp_str = y.rstrip().split()[0][0:-1] #grab line count - ':'
                if temp_str == '-':
                    continue
                if temp_str == '#####':
                    label.append(0)
                else:
                    label.append(int(temp_str.replace("*","")))
        return label
    
    def parse_output(self, filepath):
        info_f = open(filepath, "r")
        info_lines = info_f.readlines()
        i = 0
        total = 0
        error_msgs = {">I can't see one here.", ">You can't go that way.", 
                      ">The window is closed.", ">I could't find anything.", 
                      ">You are empty handed.", ">You are in perfect health.",
                     ">You cannot climb any higher."}
        for y in info_lines:
            ystr = y.rstrip()
            if len(ystr) and ystr[0] == '>':
                total += 1
                if(not (ystr in error_msgs)):
                    i += 1
        return i / total
    
    
# for when samples are each in their own subfolder, 

class zork_dataset(torch.utils.data.Dataset):
    
    def __init__(self, folder_path, low_bound, high_bound, real=1.0):
        self.path = folder_path
        self.game_ns = []
        self.step_ns = []
        self.real = real
        for i in range(low_bound,high_bound+1):
         #   self.game_ns.append(tmp_game)
            self.step_ns.append(str(i))

            #parse and store labels
            full_path = os.path.join(folder_path, "Game"+str(i))

            info_paths = []
            info_paths.append(os.path.join(full_path, "rooms.c.gcov")) # CHANGE if i want more files
            info_paths.append(os.path.join(full_path, "nrooms.c.gcov")) # CHANGE if i want more files
            label = self.get_coverage(info_paths) # grab coverage information from files
            coverage_score = 0.0
            for i in label:
                if i > 0.5:
                    coverage_score += 1.0
            coverage_denom = max(1, len(label))
            coverage_score = coverage_score / coverage_denom

            #print(coverage_score)
            output_validity = self.parse_output(os.path.join(full_path, "ZorkOut.gameOut"))
            #print(output_validity)

            full_label = []
            full_label.append(real)
            full_label.append(output_validity)
            full_label.append(coverage_score)

            label_path = os.path.join(full_path, "label.csv")
            with open(label_path, 'w', newline='') as myfile:
                wr = csv.writer(myfile)
                wr.writerow(full_label)

        self.len = len(self.step_ns)
        
    def __getitem__(self, index):
        item_path = os.path.join(self.path, "Game"+self.step_ns[index])
        input_path = os.path.join(item_path, "input")
        label_path = os.path.join(item_path, "label.csv")
        x = np.loadtxt(input_path, delimiter='\n', dtype='str')
        x = x.tolist()
        y = np.loadtxt(label_path, delimiter=',', dtype=np.float32)  
        return x, y
    
    def __len__(self):
        return self.len
    
    def get_coverage(self, filepaths):
        label = []
        for filepath in filepaths:
            info_f = open(filepath, "r")
            info_lines = info_f.readlines()
            i = 0
            for y in info_lines:
                i += 1
                temp_str = y.rstrip().split()[0][0:-1] #grab line count - ':'
                if temp_str == '-':
                    continue
                if temp_str == '#####':
                    label.append(0)
                else:
                    label.append(int(temp_str.replace("*","")))
        return label
    
    def parse_output(self, filepath):
        info_f = open(filepath, "r")
        info_lines = info_f.readlines()
        i = 0
        total = 0
        error_msgs = {">I can't see one here.", ">You can't go that way.", ">The window is closed.",
                      ">I could't find anything.", ">You are empty handed.", ">You are in perfect health.",
                     ">You cannot climb any higher."}
        for y in info_lines:
            ystr = y.rstrip()
            if len(ystr) and ystr[0] == '>':
                total += 1
                if(not (ystr in error_msgs)):
                    i += 1
        return i / total
    
    
def write_vocab_tofile(vocab_list):
    with open('zork_vocab.txt', 'w') as vocab_file:
        for word in vocab_list:
            vocab_file.write('%s\n' % word)

def my_stoi(TEXT, x):
    return TEXT.vocab.stoi[x]

def my_stoi2(vocab_list, x):
    return vocab_list.index(x)

def load_vocab_file():
    vocab_list = []
    file1 = open('zork_vocab.txt', 'r') 
    Lines = file1.readlines() 
    for line in Lines: 
        vocab_list.append(line.strip())
    return vocab_list

def save_zorkdata(path, prefix, inputs, lengths, labels):
    full_prefix = os.path.join(path, prefix)
    torch.save(inputs, full_prefix + '_inputs.pt')
    torch.save(lengths, full_prefix + '_lengths.pt')
    torch.save(labels, full_prefix + '_labels.pt')

def load_zorkdata(path, prefix):
    full_prefix = os.path.join(path, prefix)
    inputs = torch.load(full_prefix + '_inputs.pt')
    lengths = torch.load(full_prefix + '_lengths.pt')
    labels = torch.load(full_prefix + '_labels.pt')
    return inputs, lengths, labels